@@ -15,6 +15,8 @@
 #define max(a, b) ((a) > (b) ? (a) : (b))
 #define elements(a) (sizeof(a) / sizeof((a)[0]))
 
+static char *hexchars = "0123456789abcdef";
+
 typedef enum
 {
     DNS_REC_INVALID = -1, // Error code
@@ -45,7 +47,10 @@ typedef enum
     DNS_REC_NSEC = 47,
     DNS_REC_NSEC3 = 50,
     DNS_REC_NSEC3PARAM = 51,
+    DNS_REC_SMIMEA = 53,
     DNS_REC_OPENPGPKEY = 61,
+    DNS_REC_SVCB = 64,
+    DNS_REC_HTTPS = 65,
     DNS_REC_PTR = 12,
     DNS_REC_RP = 17,
     DNS_REC_RRSIG = 46,
@@ -74,6 +79,7 @@ dns_record_type dns_str_to_record_type(const char *str)
     // Performance is important here because we may want to use this when reading
     // large numbers of DNS queries from a file.
     long int code;
+    char *endptr;
 
     switch (tolower(str[0]))
     {
@@ -133,6 +139,7 @@ dns_record_type dns_str_to_record_type(const char *str)
                             {
                                 return DNS_REC_CDNSKEY;
                             }
+                            return DNS_REC_INVALID;
                         default:
                             return DNS_REC_INVALID;
                     }
@@ -194,11 +201,23 @@ dns_record_type dns_str_to_record_type(const char *str)
                     return DNS_REC_INVALID;
             }
         case 'h':
-            if (tolower(str[1]) == 'i' && tolower(str[2]) == 'p' && str[3] == 0)
+            switch (tolower(str[1]))
             {
-                return DNS_REC_HIP;
+                case 'i':
+                    if(tolower(str[2]) == 'p' && str[3] == 0)
+                    {
+                        return DNS_REC_HIP;
+                    }
+                    return DNS_REC_INVALID;
+                case 't':
+                    if(tolower(str[2]) == 't' && tolower(str[3]) == 'p' && tolower(str[4]) == 's' && str[5] == 0)
+                    {
+                        return DNS_REC_HTTPS;
+                    }
+                    return DNS_REC_INVALID;
+                default:
+                    return DNS_REC_INVALID;
             }
-            return DNS_REC_INVALID;
         case 'i':
             if (tolower(str[1]) == 'p' && tolower(str[2]) == 's' && tolower(str[3]) == 'e' && tolower(str[4]) == 'c'
                 && tolower(str[5]) == 'k' && tolower(str[6]) == 'e' && tolower(str[7]) == 'y' && str[8] == 0)
@@ -320,6 +339,12 @@ dns_record_type dns_str_to_record_type(const char *str)
                         return DNS_REC_SIG;
                     }
                     return DNS_REC_INVALID;
+                case 'm':
+                    if (tolower(str[2]) == 'i' && tolower(str[3]) == 'm' && tolower(str[4]) == 'e' && tolower(str[5]) == 'a' && str[6] == 0)
+                    {
+                        return DNS_REC_SMIMEA;
+                    }
+                    return DNS_REC_INVALID;
                 case 'o':
                     if (tolower(str[2]) == 'a' && tolower(str[3]) == 0)
                     {
@@ -338,6 +363,12 @@ dns_record_type dns_str_to_record_type(const char *str)
                         return DNS_REC_SSHFP;
                     }
                     return DNS_REC_INVALID;
+                case 'v':
+                    if (tolower(str[2]) == 'c' && tolower(str[3]) == 'b' && str[4] == 0)
+                    {
+                        return DNS_REC_SVCB;
+                    }
+                    return DNS_REC_INVALID;
                 default:
                     return DNS_REC_INVALID;
             }
@@ -400,8 +431,8 @@ dns_record_type dns_str_to_record_type(const char *str)
         case '8':
         case '9':
             errno = 0;
-            code = strtol(str, NULL, 10);
-            if(code < 0 || code > 0xFFFF || errno != 0)
+            code = strtol(str, &endptr, 10);
+            if(code < 0 || code > 0xFFFF || errno != 0 || *endptr != 0)
             {
                 return DNS_REC_INVALID;
             }
@@ -422,7 +453,7 @@ typedef enum
 
 typedef enum
 {
-    DNS_RCODE_OK = 0,
+    DNS_RCODE_NOERROR = 0,
     DNS_RCODE_FORMERR = 1,
     DNS_RCODE_SERVFAIL = 2,
     DNS_RCODE_NXDOMAIN = 3,
@@ -446,7 +477,7 @@ bool dns_str2rcode(char *str, dns_rcode *code)
 {
     if(strcasecmp(str, "ok") == 0 || strcasecmp(str, "noerror") == 0)
     {
-        *code = DNS_RCODE_OK;
+        *code = DNS_RCODE_NOERROR;
         return true;
     }
     else if(strcasecmp(str, "formerr") == 0)
@@ -790,7 +821,7 @@ char *dns_rcode2str(dns_rcode rcode)
 
     switch (rcode)
     {
-        case DNS_RCODE_OK:
+        case DNS_RCODE_NOERROR:
             return "NOERROR";
         case DNS_RCODE_FORMERR:
             return "FORMERR";
@@ -900,6 +931,8 @@ char *dns_record_type2str(dns_record_type type)
             return "RP";
         case DNS_REC_SIG:
             return "SIG";
+        case DNS_REC_SMIMEA:
+            return "SMIMEA";
         case DNS_REC_SOA:
             return "SOA";
         case DNS_REC_SRV:
@@ -918,6 +951,10 @@ char *dns_record_type2str(dns_record_type type)
             return "TXT";
         case DNS_REC_URI:
             return "URI";
+        case DNS_REC_SVCB:
+            return "SVCB";
+        case DNS_REC_HTTPS:
+            return "HTTPS";
         default:
             snprintf(numbuf, sizeof(numbuf), "%" PRIu16, (uint16_t)type);
             return numbuf;
@@ -957,6 +994,10 @@ ssize_t dns_str2namebuf(const char *name, uint8_t *buffer)
                 *(bufname++) = 0;
                 break;
             }
+            if (*name == '.')
+            {
+                return -1;
+            }
             lenptr = bufname++;
             label_len = 0;
         }
@@ -979,17 +1020,25 @@ ssize_t dns_str2namebuf(const char *name, uint8_t *buffer)
     return total_len + 1;
 }
 
-ssize_t dns_question_create_from_name(uint8_t *buffer, dns_name_t *name, dns_record_type type, uint16_t id)
+uint16_t dns_question_size(dns_name_t *name)
+{
+    return name->length + 16;
+}
+
+uint16_t dns_question_create_from_name(uint8_t *buffer, dns_name_t *name, dns_record_type type, uint16_t id)
 {
     static uint8_t *aftername;
 
-    memcpy(buffer + 12, name->name, name->length);
-    aftername = buffer + 12 + name->length;
     dns_buffer_set_id(buffer, id);
     *((uint16_t *) (buffer + 2)) = 0;
+    *((uint16_t *) (buffer + 4)) = htons(1);
+    *((uint16_t *) (buffer + 6)) = 0;
+    *((uint16_t *) (buffer + 8)) = 0;
+    *((uint16_t *) (buffer + 10)) = 0;
+    memcpy(buffer + 12, name->name, name->length);
+    aftername = buffer + 12 + name->length;
     *((uint16_t *) aftername) = htons(type);
     *((uint16_t *) (aftername + 2)) = htons(DNS_CLS_IN);
-    *((uint16_t *) (buffer + 4)) = htons(0x0001);
     return aftername + 4 - buffer;
 }
 
@@ -1275,14 +1324,14 @@ bool dns_create_reply(uint8_t *buffer, size_t *len, char *name, dns_record_type
     return true;
 }
 
-bool dns_print_readable(char **buf, size_t buflen, const uint8_t *source, size_t len)
+bool dns_print_readable(char **buf, size_t buflen, const uint8_t *source, size_t len, bool is_name)
 {
     char *endbuf = *buf + buflen;
     size_t label_length_offset = 0;
 
     for(size_t i = 0; i < len; i++)
     {
-        if(i == label_length_offset)
+        if(i == label_length_offset && is_name)
         {
             if(endbuf - *buf <= 1)
             {
@@ -1296,7 +1345,8 @@ bool dns_print_readable(char **buf, size_t buflen, const uint8_t *source, size_t
             label_length_offset += source[i] + 1;
             continue;
         }
-        if(source[i] >= ' ' && source[i] <= '~' && source[i] != '\\' && source[i] != '.')
+        if(source[i] >= ' ' && source[i] <= '~' && source[i] != '\\' && (source[i] != '.'
+            || (source[i] == '.' && !is_name)))
         {
             if(endbuf - *buf <= 1)
             {
@@ -1329,7 +1379,7 @@ char* dns_name2str(dns_name_t *name)
     static char buf[0xFF * 4];
 
     char *ptr = buf;
-    dns_print_readable(&ptr, sizeof(buf), name->name, name->length);
+    dns_print_readable(&ptr, sizeof(buf), name->name, name->length, true);
     return buf;
 }
 
@@ -1341,12 +1391,14 @@ void dns_question2str(dns_question_t *question, char *buf, size_t len)
              dns_record_type2str(question->type));
 }
 
-char* dns_raw_record_data2str(dns_record_t *record, uint8_t *begin, uint8_t *end)
+char* dns_raw_record_data2str(dns_record_t *record, uint8_t *begin, uint8_t *end, bool put_quotes)
 {
     static char buf[0xFFFF0];
     static dns_name_t name;
+    int written;
 
     char *ptr = buf;
+    char *buf_end = buf + sizeof(buf);
 
     switch(record->type)
     {
@@ -1355,7 +1407,7 @@ char* dns_raw_record_data2str(dns_record_t *record, uint8_t *begin, uint8_t *end
         case DNS_REC_DNAME:
         case DNS_REC_PTR:
             parse_name(begin, record->data.raw, end, name.name, &name.length, NULL);
-            dns_print_readable(&ptr, sizeof(buf), name.name, name.length);
+            dns_print_readable(&ptr, buf_end - ptr, name.name, name.length, true);
             break;
         case DNS_REC_MX:
             if(record->length < 3)
@@ -1365,7 +1417,7 @@ char* dns_raw_record_data2str(dns_record_t *record, uint8_t *begin, uint8_t *end
             parse_name(begin, record->data.raw + 2, end, name.name, &name.length, NULL);
             int no = sprintf(buf, "%" PRIu16 " ", ntohs(*((uint16_t*)record->data.raw)));
             ptr += no;
-            dns_print_readable(&ptr, sizeof(buf), name.name, name.length);
+            dns_print_readable(&ptr, buf_end - ptr, name.name, name.length, true);
             break;
         case DNS_REC_TXT:
         {
@@ -1376,11 +1428,13 @@ char* dns_raw_record_data2str(dns_record_t *record, uint8_t *begin, uint8_t *end
                 uint8_t length = *(data_ptr++);
                 if (data_ptr + length <= record_end)
                 {
-                    *(ptr++) = '"';
-                    dns_print_readable(&ptr, sizeof(buf), data_ptr, length);
+                    if (put_quotes) *(ptr++) = '"';
+                    dns_print_readable(&ptr, buf_end - ptr, data_ptr, length, false);
                     data_ptr += length;
-                    *(ptr++) = '"';
-                    *(ptr++) = ' ';
+                    if (put_quotes){
+                        *(ptr++) = '"';
+                        *(ptr++) = ' ';
+                    }
                 }
                 else
                 {
@@ -1400,7 +1454,7 @@ char* dns_raw_record_data2str(dns_record_t *record, uint8_t *begin, uint8_t *end
             }
 
             parse_name(begin, record->data.raw, end, name.name, &name.length, &next);
-            dns_print_readable(&ptr, sizeof(buf), name.name, name.length);
+            dns_print_readable(&ptr, buf_end - ptr, name.name, name.length, true);
             *(ptr++) = ' ';
 
             if(next + 20 >= record->data.raw + record->length)
@@ -1408,7 +1462,7 @@ char* dns_raw_record_data2str(dns_record_t *record, uint8_t *begin, uint8_t *end
                 goto raw;
             }
             parse_name(begin, next, end, name.name, &name.length, &next);
-            dns_print_readable(&ptr, sizeof(buf), name.name, name.length);
+            dns_print_readable(&ptr, buf_end - ptr, name.name, name.length, true);
             *(ptr++) = ' ';
             if(next + 20 > record->data.raw + record->length)
             {
@@ -1443,23 +1497,61 @@ char* dns_raw_record_data2str(dns_record_t *record, uint8_t *begin, uint8_t *end
             {
                 goto raw;
             }
-            int written = sprintf(ptr, "%" PRIu8 " ", (uint8_t)(record->data.raw[0] >> 7));
+            written = sprintf(ptr, "%" PRIu8 " ", (uint8_t)(record->data.raw[0] >> 7));
             if(written < 0)
             {
                 return buf;
             }
             ptr += written;
-            dns_print_readable(&ptr, sizeof(buf), record->data.raw + 2, record->data.raw[1]);
+            dns_print_readable(&ptr, buf_end - ptr, record->data.raw + 2, record->data.raw[1], false);
             *(ptr++) = ' ';
             *(ptr++) = '"';
-            dns_print_readable(&ptr, sizeof(buf), record->data.raw + 2 + record->data.raw[1],
-                               (size_t)(record->length - record->data.raw[1] - 2));
+            dns_print_readable(&ptr, buf_end - ptr, record->data.raw + 2 + record->data.raw[1],
+                               (size_t) (record->length - record->data.raw[1] - 2), false);
             *(ptr++) = '"';
             *ptr = 0;
             break;
+        case DNS_REC_SRV:
+            if(record->length < 7)
+            {
+                goto raw;
+            }
+            uint16_t prio = ntohs(*((uint16_t*)(record->data.raw + 0)));
+            uint16_t weight = ntohs(*((uint16_t*)(record->data.raw + 2)));
+            uint16_t port = ntohs(*((uint16_t*)(record->data.raw + 4)));
+            written = snprintf(ptr, buf_end - ptr, "%" PRIu16 " %" PRIu16 " %" PRIu16 " ", prio, weight, port);
+            if(written < 0)
+            {
+                return buf;
+            }
+            ptr += written;
+            parse_name(begin, record->data.raw + 6, end, name.name, &name.length, NULL);
+            dns_print_readable(&ptr, buf_end - ptr, name.name, name.length, true);
+            break;
+        case DNS_REC_SMIMEA:
+        case DNS_REC_TLSA:
+            if(record->length < 3)
+            {
+                goto raw;
+            }
+            written = snprintf(ptr, buf_end - ptr, "%" PRIu8 " %" PRIu8 " %" PRIu8 " ",
+                               record->data.raw[0], record->data.raw[1], record->data.raw[2]);
+            if(written < 0)
+            {
+                return buf;
+            }
+            ptr += written;
+
+            for(uint16_t i = 3; i < record->length; i++) {
+                uint8_t byte = record->data.raw[i];
+                *(ptr++) = hexchars[byte >> 4];
+                *(ptr++) = hexchars[byte & 0xF];
+            }
+            *ptr = 0;
+            break;
         raw:
         default:
-            dns_print_readable(&ptr, sizeof(buf), record->data.raw, record->length);
+            dns_print_readable(&ptr, sizeof(buf), record->data.raw, record->length, false);
             *ptr = 0;
     }
     return buf;
@@ -1533,7 +1625,7 @@ void dns_print_packet(FILE *f, dns_pkt_t *packet, uint8_t *begin, size_t len, ui
 
     fprintf(f,
              ";; ->>HEADER<<- opcode: %s, status: %s, id: %"PRIu16"\n"
-             ";; flags: %s%s%s%s%s; QUERY: %" PRIu16 ", ANSWER: %" PRIu16 ", AUTHORITY: %" PRIu16 ", ADDITIONAL: %" PRIu16 "\n\n"
+             ";; flags: %s%s%s%s%s%s%s; QUERY: %" PRIu16 ", ANSWER: %" PRIu16 ", AUTHORITY: %" PRIu16 ", ADDITIONAL: %" PRIu16 "\n\n"
              ";; QUESTION SECTION:\n",
              dns_opcode2str((dns_opcode)packet->head.header.opcode),
              dns_rcode2str((dns_rcode)packet->head.header.rcode),
@@ -1543,6 +1635,8 @@ void dns_print_packet(FILE *f, dns_pkt_t *packet, uint8_t *begin, size_t len, ui
              packet->head.header.aa ? "aa " : "",
              packet->head.header.rd ? "rd " : "",
              packet->head.header.ra ? "ra " : "",
+             packet->head.header.tc ? "tc " : "",
+             packet->head.header.cd ? "cd " : "",
              packet->head.header.q_count,
              packet->head.header.ans_count,
              packet->head.header.auth_count,
@@ -1568,9 +1662,92 @@ void dns_print_packet(FILE *f, dns_pkt_t *packet, uint8_t *begin, size_t len, ui
                 rec.ttl,
                 dns_class2str((dns_class)rec.class),
                 dns_record_type2str((dns_record_type) rec.type),
-                dns_raw_record_data2str(&rec, begin, begin + len));
+                dns_raw_record_data2str(&rec, begin, begin + len, true));
     }
     fprintf(f, "\n\n");
 }
 
+uint8_t dns_ip_octet2label(uint8_t *dst, uint8_t octet)
+{
+    uint8_t length = octet >= 100 ? 3 : (octet >= 10 ? 2 : 1);
+    for(uint8_t i = 0; i < length; i++)
+    {
+        uint8_t digit = octet % 10;
+        dst[length - 1 - i] = (char)('0' + digit);
+        octet /= 10;
+    }
+    return length;
+}
+
+bool dns_ip2ptr(const char *qname, dns_name_t *name)
+{
+    struct sockaddr_storage ptr_addr_storage;
+
+    if (inet_pton(AF_INET, qname, &((struct sockaddr_in*)&ptr_addr_storage)->sin_addr) == 1)
+    {
+        name->length = 0;
+        uint8_t *addr = (uint8_t*)&((struct sockaddr_in*)&ptr_addr_storage)->sin_addr.s_addr;
+#define ipv4b2l(INDEX) name->name[name->length] = dns_ip_octet2label(name->name + name->length + 1, addr[(INDEX)]); \
+        name->length += name->name[name->length] + 1;
+        ipv4b2l(3);
+        ipv4b2l(2);
+        ipv4b2l(1);
+        ipv4b2l(0);
+        name->name[name->length++] = 7;
+        name->name[name->length++] = 'i';
+        name->name[name->length++] = 'n';
+        name->name[name->length++] = '-';
+        name->name[name->length++] = 'a';
+        name->name[name->length++] = 'd';
+        name->name[name->length++] = 'd';
+        name->name[name->length++] = 'r';
+        name->name[name->length++] = 4;
+        name->name[name->length++] = 'a';
+        name->name[name->length++] = 'r';
+        name->name[name->length++] = 'p';
+        name->name[name->length++] = 'a';
+        name->name[name->length++] = 0;
+
+        return true;
+    }
+    else if (inet_pton(AF_INET6, qname, &((struct sockaddr_in6*)&ptr_addr_storage)->sin6_addr) == 1)
+    {
+        uint8_t *addr = (uint8_t*)&((struct sockaddr_in6*)&ptr_addr_storage)->sin6_addr.s6_addr;
+#define ipv6b2l(INDEX) name->name[(INDEX) * 4 + 0] = 1; \
+    name->name[(INDEX) * 4 + 1] = hexchars[addr[15 - (INDEX)] & 0xF]; \
+    name->name[(INDEX) * 4 + 2] = 1; \
+    name->name[(INDEX) * 4 + 3] = hexchars[(addr[15 - (INDEX)] >> 4) & 0xF];
+
+        ipv6b2l(0);
+        ipv6b2l(1);
+        ipv6b2l(2);
+        ipv6b2l(3);
+        ipv6b2l(4);
+        ipv6b2l(5);
+        ipv6b2l(6);
+        ipv6b2l(7);
+        ipv6b2l(8);
+        ipv6b2l(9);
+        ipv6b2l(10);
+        ipv6b2l(11);
+        ipv6b2l(12);
+        ipv6b2l(13);
+        ipv6b2l(14);
+        ipv6b2l(15);
+        name->name[64] = 3;
+        name->name[65] = 'i';
+        name->name[66] = 'p';
+        name->name[67] = '6';
+        name->name[68] = 4;
+        name->name[69] = 'a';
+        name->name[70] = 'r';
+        name->name[71] = 'p';
+        name->name[72] = 'a';
+        name->name[73] = 0;
+        name->length = 74;
+        return true;
+    }
+    return false;
+}
+
 #endif //MASSRESOLVER_DNS_H
